# a wrapper for division script which adds support for floating point numbers

from lib.prettytable import PrettyTable
from divide import divide, table_to_text
import bitutils as bu

def get_reference_register_size(*numbers):
    return max(map(len, numbers))

def parse_float(number):
    split_by_dot = number.split('.')
    Sn = split_by_dot[0] # sign

    split_by_comma = split_by_dot[1].split(',')
    Pn = len(split_by_comma[0].lstrip('0'))
    Mn = ''.join(split_by_comma).lstrip('0')

    return Sn, Pn, Mn

def to_int(number):
    return int("0b" + number, 2)

def normalize_mantice(m, n):
    M_norm = m.lstrip('0')
    P_delta = len(m) - len(M_norm)

    print(f"Normalize: {m} -> {M_norm} (shifted right {P_delta} times)")

    return M_norm[:n], P_delta

def round_mantice(m, n):
    closest_upper = bu.sum(m[:n+1], '1', n+1)
    return closest_upper[:n]

def print_classic_float(label, Sn, Pn, Mn):
    pt = PrettyTable()
    pt.field_names = [f"S{label}", f"P{label}", f"M{label}"]
    pt.add_row([Sn, bin(Pn).lstrip('-')[2:], Mn])
    print(pt)

def print_shortened_float(label, Sn, Pn, Mn):
    mutable_mantice = list(Mn)
    P = Pn
    while P < 0:
        mutable_mantice.insert(0, '0')
        P += 1
    
    if P > len(mutable_mantice):
        for i in range(len(mutable_mantice), P+1):
            mutable_mantice.insert(i, '0')
    mutable_mantice.insert(P, ',')
    result_string = ''.join(mutable_mantice)
    print(f"{Sn}.{result_string}".replace(".,", ".0,"))

def divide_float(x, y, method, n = 0, verbose = False):
    Sx, Px, Mx = parse_float(x)
    Sy, Py, My = parse_float(y)

    print(f"Число X:\n" \
          f"Знак мантиси: {Sx}\n" \
          f"Порядок: {Px}\n" \
          f"Мантиса: {Mx}\n")

    print(f"Число Y:\n" \
          f"Знак мантиси: {Sy}\n" \
          f"Порядок: {Py}\n" \
          f"Мантиса: {My}")

    '''
    while to_int(Mx) >= to_int(My):
        print(f"Mx >= My ({Mx} >= {My}), тому зсуваємо Mx на один розряд праворуч і збільшуємо порядок числа X")
        Mx = '0' + Mx
        Px += 1
    '''

    Mx = Mx.rstrip("0")
    My = My.rstrip("0")

    print("Запис чисел у класичному форматі:")
    print_classic_float('x', Sx, Px, Mx)
    print()
    print_classic_float('y', Sy, Py, My)
    print()

    #reg_size = get_reference_register_size(Mx, My)

    while to_int(Mx) >= to_int(My):
        print(f"Mx >= My ({Mx} >= {My}), тому зсуваємо Mx на один розряд праворуч і збільшуємо порядок числа X")
        Mx = '0' + Mx
        My = My + '0'
        Px += 1

    if n == 0:
        reg_size = get_reference_register_size(Mx, My)
        print(f"n = 0, отже довжина регістрів обирається автоматично ({reg_size})")
    else:
        reg_size = n

    table, result = divide(reg_size, to_int(bu.al(Mx, reg_size)), to_int(bu.al(My, reg_size)), method)
    print(f"Процес ділення методом #{method}:\n{table_to_text(table)}\n")

    print("Маємо результат:")

    S_result = bu.xor(Sx, Sy)
    print(f"Знаковий розряд: {Sx} ⊕ {Sy} = {S_result}")

    M_norm, P_delta = normalize_mantice(result, n+1)
    print(f"Нормалізована мантиса: ,{M_norm}")

    P_result = Px - Py - P_delta
    print(f"Порядок: {Px} + {Py} - {P_delta} = {P_result}")

    M_result = round_mantice(bu.al(result, n+1), n+1)
    print(f"Округлюємо мантису до {reg_size} розрядів: ,{M_result}")

    return S_result, P_result, M_result

if __name__ == "__main__":
    start_x = input("X: ")
    start_y = input("Y: ")
    n = int(input("n: "))
    method = int(input("Method: "))

    S_result, P_result, M_result = divide_float(start_x, start_y, method, n)

    print(f"Запишемо результат у вигляді таблиці:")
    print_classic_float('f', S_result, P_result, M_result)
    print("І отримуємо остаточне число після ділення:")
    print_shortened_float('F', S_result, P_result, M_result)
